import os
import glob
import csv
import warnings
import sys
import gzip
from itertools import combinations
from shutil import copyfile, copyfileobj
from Bio.Blast.Applications import NcbimakeblastdbCommandline
from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.Sequencing.Applications import SamtoolsFaidxCommandline
from Bio.Application import ApplicationError
from Bio import SearchIO
import pybedtools as bedtools
try:
    from Magphi.exit_with_error import exit_with_error
except ModuleNotFoundError:
    from exit_with_error import exit_with_error
EXIT_INPUT_FILE_ERROR = 1

try:
    from Magphi.split_gff_file import split_single_gff
except ModuleNotFoundError:
    from split_gff_file import split_single_gff
# pylint: disable=E1123


def blast_insertion_site(seeds, genome_file, tmp_name):
    """
    Function to BLAST seed sequences against a given genome
    :param seeds: Filepath to seed sequence file
    :param genome_file: Filepath for the input genome
    :param tmp_name: The temporary name for the BLAST database constructed from the genome
    :return: Filepath to the BLAST output file in xml format
    """
    # Construct genome_db path and name
    genome_db = f'{genome_file}_tmp_db'
    # Make blast database command for given genome
    c_line_makedb = NcbimakeblastdbCommandline(dbtype='nucl', input_file=genome_file, out=genome_db)

    # Run makeblastdb in command line
    c_line_makedb()

    blast_out_xml = f'{tmp_name}.xml'
    c_line = NcbiblastnCommandline(query=seeds,
                                   db=genome_db,
                                   evalue=0.001,
                                   outfmt=5,
                                   qcov_hsp_perc=50,
                                   out=blast_out_xml)

    # Run the blast command in commandline blast
    c_line()

    # Delete blast database
    file_list = glob.glob(f'{genome_file}_tmp_db.*')
    for file in file_list:
        os.remove(file)

    return blast_out_xml


def blast_out_to_sorted_bed(blast_xml_output, include_seeds, genome_name, seed_pairs):
    """
    Function to pass BLAST output in xml format and construct BED files for each set of seed sequence pairs.
    :param blast_xml_output: File from blast in xml format for one genome queried for all seed sequence.
    :param include_seeds:
    :param genome_name:
    :param seed_pairs:
    :return:
    """
    # Read in the blast output into a generator object
    blast_output = SearchIO.parse(blast_xml_output, format='blast-xml')

    # Construct list to hold lines for seed bed files
    bed_list = []
    for qresult in blast_output:
        bed_line = [None]*4
        # Insert the name of the seed
        bed_line[3] = qresult.id
        for hit in qresult:
            bed_line_hit = bed_line.copy()
            # Get the contig that has been hit by the seed
            bed_line_hit[0] = hit.id
            for hsp in hit:
                bed_line_hsp = bed_line_hit.copy()
                # Extract the beginning of the seed (extract on to make 0-based for BED file)
                bed_line_hsp[1] = hsp.hit_start
                # Extract the end of the seed (extract on to make 0-based for BED file)
                bed_line_hsp[2] = hsp.hit_end

                # Add the bed line for the hit to the bed file list
                bed_list.append(bed_line_hsp)

    # Construct lists to hold bed file names, and a dict to hold the number of hits for each seed pair
    blast_hit_beds = []
    exclusion_list = []
    seed_hits = {}

    # Split up the bed lines into each seed and write a bed file for each seed
    for seed_pair in seed_pairs.keys():
        # Find bed lines for seed pair
        seed_bed_lines = [line for line in bed_list
                            if seed_pairs[seed_pair][0] == line[3] or seed_pairs[seed_pair][1] == line[3]]

        # Record the number of hits in a genome by a set of seeds
        seed_hits[seed_pair] = len(seed_bed_lines)

        # Write the bed files for each seed pair
        bed_name = f'{genome_name}~~{seed_pair}.bed'
        with open(bed_name, 'w') as bed_file:
            writer = csv.writer(bed_file, delimiter='\t')
            for line in seed_bed_lines:
                writer.writerow(line)
        bed_file.close()
        blast_hit_beds.append(bed_name)

        # Sort the bed file
        sorted_bed = bedtools.BedTool(bed_name).sort()
        sorted_bed.saveas(bed_name)

        # Check if seeds are to be excluded, if then copy seed bed to be used later in bedtools intersect
        if not include_seeds:
            copyfile(f'{genome_name}~~{seed_pair}.bed', f'{genome_name}~~{seed_pair}_seeds.bed')
            exclusion_list.append(f'{genome_name}~~{seed_pair}_seeds.bed')

    return blast_hit_beds, exclusion_list, seed_hits


def write_bed_from_list_of_seeds(list_of_seeds, bed_file_name):
    """
    Function to write BED file from location of seed sequences to be used when subtracting seeds from regions of interest
    :param list_of_seeds: List of seeds and their information on coordinate
    :param bed_file_name: name of the BED file to be constructed
    :return: None
    """
    # Write out new .bed file with only the extended seeds.
    # Join the items in the lists (seeds) into a sting
    list_of_seeds = [' '.join(seed) for seed in list_of_seeds]

    # Join the seeds into one large sting
    list_of_seeds = '\n'.join(list_of_seeds)

    # Construct BedTools object
    bed_of_seeds = bedtools.BedTool(list_of_seeds, from_string=True)

    # Sort the BedTools object
    bed_of_seeds = bed_of_seeds.sort()

    # Save the seeds into a .bed file
    bed_of_seeds.saveas(bed_file_name)


def seed_reach_contig_end_calc(genome_file, max_seed_dist, seed_contig_hits):
    """
    Function to examine a set of seed sequences and their ability to reach the end of contigs on which they are placed.
    :param genome_file: Filepath to the genome in question
    :param max_seed_dist: Maximum distance allowed between seed sequences
    :param seed_contig_hits: dict with each key being a contig and values being a list of seed sequence hit on contigs as lists

    :return end_reaches: List with two indexes of seed sequences that reach either 2 or 1 end of their respective contig
    :return end_sums: List with two indexes of the number of times seed sequences reach the 5' or 3' end
    :return end_reached_matrix: Matrix (list of lists) that is NxN (n = number of seed sequences being examined) index is 1 when primers can interact across contig breaks
    :return intervals: List of the coordinate intervals in which seed sequences are located.
    """
    # Construct dict to hold the indexing of the fasta file
    fai_dict = {}
    # Read and search the .fai file to get the edge of contigs
    with open(genome_file, 'r') as fai_file:
        for line in fai_file.readlines():
            line = line.split('\t')
            fai_dict[line[0]] = int(line[1])
    fai_file.close()

    # Extract all lines from dict holding Bed intervals for seed hits
    intervals = [interval for key in seed_contig_hits.keys() for interval in seed_contig_hits[key]]

    # Insert line identifier for interaction matrix
    [line.append(str(i)) for i, line in enumerate(intervals)]

    # Construct matrix to hold info on contig ends reached
    end_reached_matrix = [[0] * len(intervals) for _ in range(len(intervals))]

    # Go through each seed hit to see if it reached the end of the contig under the max distance
    for i, interval in enumerate(intervals):
        # Check if the 5' and 3' can be reached using the max distance
        if int(interval[1]) - max_seed_dist < 0:
            end_reached_matrix[i][0] = 1

        if int(interval[2]) + max_seed_dist > fai_dict[interval[0]]:
            end_reached_matrix[i][1] = 1

    # Find how many ends each seed can reach
    end_sums = [sum(ends) for ends in end_reached_matrix]

    # Count the number of seeds that reach one or both ends
    end_reaches = [end_sums.count(2), end_sums.count(1)]

    return end_reaches, end_sums, end_reached_matrix, intervals


def examine_flanking_regions(seed_contig_hits, max_seed_dist, genome_file, bed_file_name=None):
    '''
    Function that takes a set of seeds on contigs, and search for how they can be connected using a genome and length of contigs.
    :param seed_contig_hits: dict with each key being a contig and values being a list of seed sequence hit on contigs as lists
    :param max_seed_dist: int for the maximum distance allowed between seed sequenes
    :param genome_file: .fai (fasta-index) of a genome used as info on contig sizes
    :param bed_file_name: Used for saving updated BED file with coordinates
    :return: Evidence level for a set of seed seuqences that have hit the genome.
    '''
    # If no limits are given and only one contig is hit then no pairs can be predicted
    if max_seed_dist == 0:
        return 2

    # If limits are given then check overlap of seeds, if they are only found on one contig
    elif len(seed_contig_hits) == 1 and max_seed_dist > 0:
        # Extract the intervals found for seeds
        intervals = seed_contig_hits[list(seed_contig_hits.keys())[0]]

        # Construct an interaction matrix for the seeds
        interaction_matrix = [[0]*len(intervals) for _ in range(len(intervals))]

        # Insert line identifier for interaction matrix
        [line.append(str(i)) for i, line in enumerate(intervals)]

        for i, interval in enumerate(intervals):
            # temp intervals to be altered
            intervals_tmp = intervals.copy()

            # Convert interval from list to BedTools object
            bed_interval = ' '.join(interval)
            bed_interval = bedtools.BedTool(bed_interval, from_string=True)

            # slop on the object with the max distance on both sides
            bed_interval = bed_interval.slop(b=max_seed_dist, g=genome_file)

            # Reinsert the newly slopped interval
            intervals_tmp[i] = list(bed_interval[0])

            # Construct BedTool object from all seed hits in intervals
            # Join the intervals into strings
            intervals_tmp = [' '.join(hit) for hit in intervals_tmp]
            # Join the string intervals into one big string
            intervals_tmp = '\n'.join(intervals_tmp)

            # Create a BedTool object from the large string
            intervals_bed = bedtools.BedTool(intervals_tmp, from_string=True)

            # Resort if needed
            intervals_bed = intervals_bed.sort()

            # merge
            merged_intervals = intervals_bed.merge(d=0, c=[4, 4, 5], o='count,collapse,collapse')

            # Check to see if some lines have been merged, if then record this in the interaction matrix
            interacting_lines = [list(line) for line in merged_intervals if int(line[3]) >= 2]

            # Filter out overlapping seeds in which the seeds are the same
            interacting_ids = [line[5] for line in interacting_lines if len(set(line[4].split(','))) >= 2]
            interacting_seed_names = [line[4] for line in interacting_lines if len(set(line[4].split(','))) >= 2]

            # Insert interacting seeds in matrix (only in upper part of matrix):
            if len(interacting_ids) > 0:
                for cluster in zip(interacting_ids, interacting_seed_names):
                    # split the pairs in the zipped tuple and find the combinations of names and ids
                    id_combinations = combinations(cluster[0].split(','), 2)
                    name_combinations = combinations(cluster[1].split(','), 2)

                    # Go through each pair of id and name combination to record valid hits
                    for pair in zip(id_combinations, name_combinations):
                        # Check that the combination of seeds are not the same seed and that seed being examined is involved
                        if pair[1][0] != pair[1][1] and str(i) in pair[0]:
                            interaction_matrix[int(pair[0][0])][int(pair[0][1])] = 1


        # Summarise all rows of the interaction matrix
        num_interactions = [sum(row) for row in interaction_matrix]
        # Summarise col values to get a number of total interactions
        num_interactions = sum(num_interactions)

        if num_interactions == 0:
            return 1

        elif num_interactions == 1:
            # Check if seeds reach end of contigs
            end_reaches, *_ = seed_reach_contig_end_calc(genome_file, max_seed_dist, seed_contig_hits)
            if sum(end_reaches) > 0:
                return 2

            interaction_indexes = []
            # Get index of the interaction between seeds
            for i, interacts in enumerate(interaction_matrix):
                if 1 in interacts:
                    for j, interaction in enumerate(interaction_matrix[i]):
                        if interaction == 1:
                            interaction_indexes = [i, j]

            # Extract the seeds that correspond to the identified indexes
            interaction_seeds = [intervals[index][:4] for index in interaction_indexes]
            # Write the bed file from the list of seeds
            write_bed_from_list_of_seeds(interaction_seeds, bed_file_name)
            return '5B'

        elif num_interactions > 1:
            return 2

    else:
        end_reaches, end_sums, end_reached_matrix, intervals = seed_reach_contig_end_calc(genome_file, max_seed_dist, seed_contig_hits)

        # See if no ends has been reached
        if end_reaches[0] == 0 and end_reaches[1] == 0:
            return 1

        # Check if one seed has reached one or both ends, while the other has reached no ends.
        elif (end_reaches[0] == 1 and end_reaches[1] == 0) or (end_reaches[0] == 0 and end_reaches[1] == 1):
            return 3

        # Check if
        # two or more seeds hit both ends of their contig (end_reaches[0] > 1) or
        # three or more seeds all hit one end of their contig (end_reaches[1] > 2) or
        # one seed reaches two ends and two seeds reaches one end each (end_reaches[0] == 1 and end_reaches[1] == 2)
        elif end_reaches[0] > 1 or end_reaches[1] > 2 or (end_reaches[0] == 1 and end_reaches[1] == 2):
            return 2

        # Check if one seed has reached both ends and one seed has reached one end
        elif end_reaches[1] == 1 and end_reaches[0] == 1:
            return 2

        # Check if two seeds have reached one end.
        elif end_reaches[1] == 2:
            # Organise the bed and reprint with the usefull seeds and adjusted coordinates to end
            seed_indexes = [index for index, reached in enumerate(end_sums) if reached == 1]

            # Initialise list to hold extended seeds
            extended_seeds = []

            for index in seed_indexes:
                # Get the seed extension hits
                seed_extension_hits = end_reached_matrix[index]

                # Get the seed in question, and remove the identifier at the end.
                extend_seed = intervals[index][:4]

                # Convert seed to a BedTools object
                extend_seed = bedtools.BedTool(' '.join(extend_seed), from_string=True)

                # Examine if seed should be extended in 5' end
                if seed_extension_hits[0] == 1:
                    extend_seed = extend_seed.slop(g=genome_file, l=max_seed_dist, r=0)

                # Examine if the seed should be extended in the 3' end
                elif seed_extension_hits[1] == 1:
                    extend_seed = extend_seed.slop(g=genome_file, r=max_seed_dist, l=0)

                else:
                    warnings.warn("A seed found to reach the end of a contig's "
                                  "end was not extended in any direction!\n"
                                  f"Please report this along with the genome: {genome_file},"
                                  f" the seed in question: {extend_seed}, "
                                  f"and the max distance set: {max_seed_dist}")
                    return "Something unexpected was found, please check warning log!"

                extended_seeds.append(extend_seed[0])

            # Write out new .bed file with only the extended seeds.
            write_bed_from_list_of_seeds(extended_seeds, bed_file_name)
            return '4B'

        else:
            warnings.warn(f'Some unaccounted for constalation of seeds hits '
                          f'was found to reach the end of contigs.\n'
                          f'Genome .fai file examined: {genome_file}.\n'
                          f'Please attach the .fai file (found in the tmp folder)'
                          f' when reporting this, along with the following: '
                          f'{intervals} and the max distance used: {max_seed_dist}')
            return "Something unexpected was found, please check warning log!"


def check_seeds_placement(bed_files, seed_pairs, seed_hits, max_seed_dist, genome_file, file_type, tmp_folder):
    """
    Function to determine the placement of seeds and how to search for connections between them
    :param bed_files: List of bed files containing the coordinates of seed sequences
    :param seed_pairs: Dict containing the common names of seed sequence pairs as keys and values being list with both full length seed seqenuce names
    :param seed_hits: Dict with the key being common name of seed sequence pair and the value being the number of hits in the given genome from the pair
    :param max_seed_dist: Maximum distance between seeds allowed by the user
    :param genome_file: Filepath to the genome in question
    :param file_type: String of either 'gff' or 'fasta' indicating the input file type for genomes
    :param tmp_folder: Filepath to the temporary folder for Magphi
    :return: A genome file in the temporary directory and an evidence score for each bed file and its seed connections
    """
    # Produce genome index file using samtools
    samtools_faidx_cmd = SamtoolsFaidxCommandline(ref=genome_file)
    try:
        samtools_faidx_cmd()
    except ApplicationError:
        exit_with_error(f"Samtools failed when running the command 'samtools faidx {genome_file}'", EXIT_INPUT_FILE_ERROR)


    # Initialise the dict to hold the support for a seed hit
    seed_hit_support_dict = dict.fromkeys(seed_pairs)

    # Check each BED file
    for file in bed_files.copy():
        # Strip file name to get name of genome
        seed_name = file.rsplit('.', 1)[0]
        seed_name = seed_name.rsplit('~~', 1)[-1]

        # Initialise dict to hold information on seed distribution over contigs
        seed_to_contig = {}

        # Check if more than one seeds has hit the genome,
        # if, then check if they are same contig
        if seed_hits[seed_name] > 1:
            # Read Bed file
            bed_file = bedtools.BedTool(file)

            # Index lines in BED file and which contig seed sequences hit
            for line in bed_file:
                contig = line[0]
                try:
                    # Add the line to the contig
                    seed_to_contig[contig] += [list(line)]
                except KeyError:
                    # If not found add the contig and the line
                    seed_to_contig[contig] = [list(line)]

            # Find number of different seeds that hit the contigs
            # Get all hits on contigs an unpack them into a single list of lists, then count the number of unique seed sequence hits
            seed_bed_lines = [seed_to_contig[key] for key in seed_to_contig.keys()]
            seed_bed_lines = [hit for contig in seed_bed_lines for hit in contig]
            uniq_seeds = len(set([hit[3] for hit in seed_bed_lines]))

            # If more than one contig is hit and more than one seed sequence has hit then examine further,
            # else only one contig is hit
            if len(seed_to_contig.keys()) > 1 and uniq_seeds > 1:
                # Examine if some of the seeds are on the same contig:
                multi_hit_contigs = [contig for contig in seed_to_contig.keys() if len(seed_to_contig[contig]) >= 2]
                # Check if there are any contigs hit with multiple seeds,
                # if, then examine them for overlaps between seeds,
                # if not, examine across contigs
                if len(multi_hit_contigs) > 0:
                    # Examine all contigs that contain multiple seeds for an overlap
                    for contig in multi_hit_contigs:
                        hit_contig = {contig: seed_to_contig[contig]}

                        return_value = examine_flanking_regions(hit_contig, max_seed_dist, f'{genome_file}.fai', file+'_5')

                        # Record the return value as evidence
                        # 1 or 2 means more examination is required
                        if return_value == 2:
                            seed_hit_support_dict[seed_name] = return_value
                            try:
                                bed_files.remove(file)
                            except ValueError:
                                pass
                        elif return_value == 1:
                            # Examine the remaining junctions to see if two seeds can be found to connect across contigs
                            seed_hit_support_dict[seed_name] = examine_flanking_regions(seed_to_contig,
                                                                                            max_seed_dist,
                                                                                            f'{genome_file}.fai',
                                                                                            file)
                            # Check if no connection could be made on or across contigs, if then delete bed from further processing
                            if seed_hit_support_dict[seed_name] == 1 or seed_hit_support_dict[seed_name] == 2:
                                bed_files.remove(file)

                        elif return_value == '5B':
                            seed_hit_support_dict[seed_name] = return_value
                            alternative_return_value = examine_flanking_regions(seed_to_contig,
                                                                                max_seed_dist,
                                                                                f'{genome_file}.fai',
                                                                                file+'_6')

                            if alternative_return_value == '4B': # ADD alternative value 3?
                                seed_hit_support_dict[seed_name] = 2
                                os.remove(file+'_5')
                                try:
                                    os.remove(file + '_6')
                                except FileNotFoundError:
                                    pass

                            else:
                                seed_hit_support_dict[seed_name] = return_value
                                os.rename(file+'_5', file)

                        else:
                            sys.exit(100)
                else:
                    # Examine the seed hits across different contigs
                    return_value = examine_flanking_regions(seed_to_contig,
                                                            max_seed_dist,
                                                            f'{genome_file}.fai',
                                                            file)
                    if return_value == 1 and seed_hits[seed_name] == 2 and uniq_seeds == 2:
                        seed_hit_support_dict[seed_name] = '4A'
                    elif return_value == 2:
                        seed_hit_support_dict[seed_name] = return_value
                        bed_files.remove(file)
                    else:
                        seed_hit_support_dict[seed_name] = return_value

            else:
                # Check if precisely two seeds hit the contig and that the mates from the pair hit one time each
                if seed_hits[seed_name] == 2 and uniq_seeds == 2:
                    # Test if seeds can reach end of contig
                    return_value = examine_flanking_regions(seed_to_contig, max_seed_dist, f'{genome_file}.fai')
                    # Score seed hit
                    if return_value == 1:
                        seed_hit_support_dict[seed_name] = '5A'
                    else:
                        seed_hit_support_dict[seed_name] = return_value

                # Check if two unique seeds hit, but one/both may have hit multiple times
                elif seed_hits[seed_name] > 2 and uniq_seeds == 2:
                    # Examine seed hits and record returned score
                    seed_hit_support_dict[seed_name] = examine_flanking_regions(seed_to_contig,
                                                                                    max_seed_dist,
                                                                                    f'{genome_file}.fai')
                    if seed_hit_support_dict[seed_name] == 1 or seed_hit_support_dict[seed_name] == 2:
                        bed_files.remove(file)

                # Check that only one seed has hit, but it has hit multiple times
                elif uniq_seeds == 1:
                    # Score seed hit
                    seed_hit_support_dict[seed_name] = 0
                else:
                    raise NotImplementedError(f'Some unaccounted for constalation of seeds hits '
                                              f'was found to hit a single contig.\n'
                                              f'The seed pair in question is {seed_name} in genome {genome_file}.\n'
                                              f'Please report this along with the following: {seed_bed_lines}')
        else:
            # set seed evidence and remove bed-file from further processing
            seed_hit_support_dict[seed_name] = 0
            bed_files.remove(file)

    return genome_file, seed_hit_support_dict


def bed_merge_handling(blast_hit_beds, include_seeds, exclude_seed_list, max_seed_dist, seed_evidence):
    """
    Function to merge lines (seed sequence hits) in a BED file into single lines (coordinate intervals)
    :param blast_hit_beds: List of BED files to be merged
    :param include_seeds: Bool indicating if seed sequence coordinates should be included in the merged interval.
    :param exclude_seed_list: List of paths to BED files with coordinates of the seed sequences
    :param max_seed_dist: Maximum distance allowed between seed sequences
    :param seed_evidence: Dict containing keys being the common name of seed sequences and the value being the current evidence level of the seed sequence for the given genome

    :return merged_bed_files: List of BED files that have been merged
    :return seed_evidence: Dict containing keys being the common name of seed sequences and the value being the new evidence level of the seed sequence for the given genome
    """
    # initialise list to hold merged bed file names:
    merged_bed_files = []

    # Process the bed file for each seed pair
    for i, bed_file in enumerate(blast_hit_beds):
        seed_hits = bedtools.BedTool(bed_file)

        # Get name of seed
        split_bed_name = bed_file.rsplit('/', 1)[-1]
        split_bed_name = split_bed_name.replace('.bed', '')
        seed_name = split_bed_name.rsplit('~~', 1)[-1]

        # Check if max distance between seeds is set to unlimited (0)
        if max_seed_dist == 0:
            max_seed_dist = 9999999999

        # Merge the bed file and collapse the column containing the names and count the number of intervals collapsed
        if len(seed_hits) > 0:
            seed_hits = seed_hits.merge(c=[4, 4], o='collapse,count', d=max_seed_dist)
        else:
            seed_evidence[seed_name] = 0
            continue

        # Check if the evidence level for the seeds should increase due to seeds being merged.
        # Check that there is one line and that two seeds have been merged and not just one seed hit
        if len(seed_hits) == 1 and int(seed_hits[0][4]) == 2:
            seed_evidence[seed_name] = '5B'

        # Evaluate if seeds are to be excluded from the intervals
        if not include_seeds:
            # Load bed containing seeds to be excluded
            exclusion_bed = bedtools.BedTool(exclude_seed_list[i])

            pre_deletion_intervals = len(seed_hits)

            # Remove the seed intervals
            seed_hits = seed_hits.subtract(exclusion_bed)

            if pre_deletion_intervals > len(seed_hits) and (str(seed_evidence[seed_name]) < '3' or seed_evidence[seed_name] == '5B' or seed_evidence[seed_name] == '4B'):
                seed_evidence[seed_name] = 3

        # Save the merged intervals
        # Construct name
        merged_bed_file = bed_file.rsplit('.', 1)[0]
        merged_bed_file = f'{merged_bed_file}_merged.bed'
        # Save
        # Check if there is any interval to report,
        # else notify by setting the return evidence level
        if len(seed_hits):
            seed_hits.saveas(merged_bed_file)
            # Add to list
            merged_bed_files.append(merged_bed_file)
        else:
            if str(seed_evidence[seed_name]) < '3':
                seed_evidence[seed_name] = 3

    return merged_bed_files, seed_evidence


def extract_seqs_n_annots(merged_bed_files, file_type, genome_file, annotation_file, tmp_folder, out_path, seed_pairs,
                          seed_evidence, print_seq_out):
    """
    Function to extract the nucleotide sequence and potential genomic features from FASTA and GFF files
    :param merged_bed_files: List of filepaths to merged BED files
    :param file_type: String indicating the input genome filetype
    :param genome_file: Filepath to the genome in question
    :param annotation_file: Filepath to a file containing annotations (None if input is fasta format)
    :param tmp_folder: Filepath to the temporary folder constructed by Magphi
    :param out_path: Filepath to the output folder
    :param seed_pairs: Dict of seed sequences that are matched into pairs
    :param seed_evidence: Dict containing keys being the common name of seed sequences and the value being the current evidence level of the seed sequence for the given genome
    :param print_seq_out: String indicating if regions spanning contig breaks should be outputted or not.

    :return annots_per_interval: Dict with keys being the common seed sequence name and values being the number of annotations between the seeds
    :return break_seeds: Dict with keys being the common seed sequence name and values being a list containing name of a seed and 'break' indicating it being next to a sequence break
    :return seed_evidence: Dict containing keys being the common name of seed sequences and the value being the new evidence level of the seed sequence for the given genome
    :return inter_seed_dist: Dict with keys being the common seed sequence name and values being the number of nucleotides between the seeds
    """
    # Initiate dict to hold number of annotations per interval
    annots_per_interval = dict.fromkeys(seed_pairs)
    # Initiate dict to hold seeds that may neighbour a sequence break
    break_seeds = {}
    # Initiate dict tot hold distances between seeds
    inter_seed_dist = {}

    # Extract sequence from Fasta from all merged bed files
    for merged_bed in merged_bed_files:
        # Read Bed file
        merged_intervals = bedtools.BedTool(merged_bed)

        # Check if the bed file is empty indicate that no intervals are present by giving nan in annotations
        # There are no intervals when excluding seeds and no seeds are merged for specific seed set.
        if file_type == 'gff':
            seed_pair = merged_bed.rsplit('~~', 1)[-1]
            seed_pair = seed_pair.rsplit('_merged', 1)[0]
            if len(merged_intervals) == 0:
                annots_per_interval[seed_pair] = 'nan'
                continue

        # Go though each line of bed, really only necessary if there are breaks in genome.
        for interval in merged_intervals:
            # Identify seed pair for interval:
            seeds = interval[3].split(',')
            # Check if seeds are merged, or if single seed interval
            if len(seeds) == 2:
                seed_pair_name = [seed_name for seed_name in seed_pairs if
                                    seeds[0] in seed_pairs[seed_name] and seeds[1] in seed_pairs[seed_name]]

                seed_pair_name = seed_pair_name[0]
            else:
                seed_pair_name = f'{seeds[0]}_break'
                break_seeds[seed_pair_name] = [seeds[0], 'break']

            # Construct output name for extracted sequence
            genome_name = genome_file.rsplit('/', 1)[-1]
            genome_name = genome_name.rsplit('.', 1)[0]
            genome_name = genome_name.rsplit('_tmp', 1)[0]
            out_file_name = f'{genome_name}--{seed_pair_name}'
            output_genome = os.path.join(out_path, f'{out_file_name}.fasta')

            # Construct good fasta header
            fasta_header = f'{genome_name}_{seed_pair_name}'

            # Convert interval to string to be made into a BedTool object
            interval = f'{interval[0]} {int(interval[1])} {int(interval[2])} {fasta_header}'

            # Create BedTool object from line
            interval = bedtools.BedTool(interval, from_string=True)

            # Record the length of the segments between seeds
            inter_seed_dist[seed_pair_name] = len(interval[0])

            # Extract the fasta sequence and save it in output folder
            if print_seq_out == 'All' or '_break' not in seed_pair_name and print_seq_out == 'output':
                try:
                    interval.sequence(fi=genome_file, fo=output_genome, nameOnly=True)
                except bedtools.helpers.BEDToolsError:
                    interval.sequence(fi=genome_file, fo=output_genome, name=True)

            # extract annotations if gff is provided as input
            if file_type == 'gff':
                # Load in the gff as a BedTool object
                gff_file = bedtools.BedTool(annotation_file)

                # Save the merged BED interval currently being handled
                interval_save = os.path.join(tmp_folder, f'interval_{genome_name}')
                interval.saveas(interval_save)

                # extract the annotations within the current merged bed interval
                annot = gff_file.intersect(b=interval, f=0.95)

                # Record the number of annotations within each interval
                annots_per_interval[seed_pair_name] = len(annot)

                # Start writing the output file, if an annotation if found
                if len(annot) > 0:
                    # Check if seed sequence reach a contig break
                    # If then get the seed pair to adjust the evidence level
                    if '_break' in seed_pair_name:
                        seed_pair_key = seed_pair_name
                        seed_pair_key = seed_pair_key.rsplit('_', 1)[0]
                        seed_pair_key = [key for key in seed_pairs.keys() if seed_pair_key in seed_pairs[key]][0]
                    else:
                        seed_pair_key = seed_pair_name

                    # Increase the evidence level for the seed sequence pair.
                    # Only if the evidence level is 4B or 5B and seed pair is valid (no breaks)
                    if seed_pair_key in seed_evidence.keys():
                        if seed_evidence[seed_pair_key] == '4B':
                            seed_evidence[seed_pair_key] = '4C'
                        elif seed_evidence[seed_pair_key] == '5B':
                            seed_evidence[seed_pair_key] = '5C'

                    # Construct name for output file
                    output_gff = os.path.join(out_path, f'{out_file_name}.gff')

                    # Find start coordinate for current merged interval
                    start_coordinate = int(interval[0][1])+1

                    # Write GFF output, if allowed or break
                    if print_seq_out == 'All' or '_break' not in seed_pair_name and print_seq_out == 'output':
                        # Open output file
                        gff_output_file = open(output_gff, 'w')
                        gff_writer = csv.writer(gff_output_file, delimiter='\t')

                        # Insert the required '##gff-version 3' line
                        gff_output_file.write('##gff-version 3\n')
                        # Write line recording the length of the extracted region
                        gff_output_file.write(f'##sequence-region {fasta_header} 1 {len(interval[0])}\n')

                        # Go through each line adjust the coordinates, the contig name and write the gff
                        for i, line in enumerate(annot):
                            line[0] = fasta_header
                            line[3] = int(line[3]) - start_coordinate
                            line[4] = int(line[4]) - start_coordinate + 1
                            gff_writer.writerow(line)

                        # Add in the tag for the Fasta
                        gff_output_file.write('##FASTA\n')
                        # Open the fasta sequence for copying
                        fasta_file = open(output_genome, 'r')
                        # Append the fasta sequence for the extracted interval
                        copyfileobj(fasta_file, gff_output_file)

                        # Close output files
                        gff_output_file.close()
                        fasta_file.close()

                # Delete the file containing the current merged bed interval
                os.remove(interval_save)

    # Remove the temporary genome and the index file (.fai), or the temporary fasta and annotation file for gff input
    if file_type == 'fasta':
        os.remove(genome_file)
        try:
            os.remove(f'{genome_file}.fai')
        except FileNotFoundError:
            pass
    else:
        try:
            os.remove(f'{genome_file}.fai')
        except FileNotFoundError:
            pass
        os.remove(genome_file)
        os.remove(annotation_file)

    return annots_per_interval, break_seeds, seed_evidence, inter_seed_dist


def screen_genome_for_seeds(genome_file, seed_pairs, seed_path, tmp_folder,
                              include_seeds, file_type, out_path, max_seed_dist, file_logger,
                              is_input_gzipped, print_seq_out):
    """
    Function that summarise the search of seeds sequences, determination of position and extraction.
    :param genome_file: Path to genome to be searched for seed sequences
    :param seed_pairs: Dict of seed sequences that are matched into pairs
    :param seed_path: Path to the file containing the seed sequence fasta
    :param tmp_folder: The working temporary folder Magphi has created
    :param include_seeds: Boolean to indicate if the coordinates of the seed sequences should be included in results
    :param file_type: Variable indicating if input genomes are gff or fasta files
    :param annotation_file: Path to the annotations part of an input gff file, previously split from the genome part
    :param out_path: Path to the output folder
    :param max_seed_dist: Integer indicating the maximum distance allowed between two seed sequences
    :param file_logger: File to with debug log statements should be written.
    :param is_input_gzipped: Bool to tell if the input file is gzipped
    :param print_seq_out: Bool to indicate if outputs related to fasta and gff sequences should be given
    :return: Number of times a seed sequence pairs hit the genome,
    number of annotations in the interval found between seed sequences,
    name of the genome extracted from,
    evidence for how seed sequences hit and connected in the genome,
    Any seed sequences that were next to a sequences break
    distance between seed sequences that could be connected.
    """
    file_logger.debug(f"Start search of sequences is {genome_file}")

    genome_name = os.path.basename(genome_file)
    genome_name = genome_name.rsplit('.', 1)[0]

    tmp_genome_folder = os.path.join(tmp_folder, genome_name)
    os.mkdir(tmp_genome_folder)

    if file_type == 'fasta':
        tmp_genome = os.path.join(tmp_genome_folder, genome_file.rsplit('/')[-1])

        if is_input_gzipped:
            file_logger.debug("\tCopying gzipped fasta file into tmp dir")
            tmp_genome = tmp_genome.split('.gz')[0]
            with gzip.open(genome_file, 'rt') as in_file:
                with open(tmp_genome, 'w') as out_file:
                    for line in in_file:
                        out_file.write(line)
        else:
            file_logger.debug("\tCopying ungzipped fasta file into tmp dir")
            copyfile(genome_file, tmp_genome)

        genome_file = tmp_genome
        annotation_file = None

    if file_type == 'gff':
        genome_file, annotation_file = split_single_gff(genome_file, tmp_genome_folder, is_input_gzipped)

    # Concatenate the genome name and the path to the temporary folder
    # genome_name = os.path.join(tmp_folder, genome_name)
    genome_name = os.path.join(tmp_genome_folder, genome_name)

    # Run blast with genome and insertion site sequences
    file_logger.debug(f"\tBLASTing: {genome_file}")
    blast_xml_output = blast_insertion_site(seed_path, genome_file, f'{genome_name}_blast')

    # Construct a bedfile from blast output
    file_logger.debug(f"\tConstructing bed file: {genome_file}")
    blast_hit_beds, exclude_seed_list, seed_hits = blast_out_to_sorted_bed(blast_xml_output,
                                                                               include_seeds,
                                                                               genome_name,
                                                                               seed_pairs)

    # Examine the seed hits and try to find solutions when multiple seeds hit at once
    file_logger.debug(f"\tChecking seed sequence placement: {genome_file}")
    genome_file, seed_evidence = check_seeds_placement(blast_hit_beds, seed_pairs, seed_hits, max_seed_dist,
                                                           genome_file, file_type, tmp_folder)

    # sort and merge the bed files
    file_logger.debug(f"\tMerging seed sequences: {genome_file}")
    merged_bed_files, seed_evidence = bed_merge_handling(blast_hit_beds, include_seeds, exclude_seed_list,
                                                           max_seed_dist, seed_evidence)

    # Extract sequences and annotations using merged intervals.
    file_logger.debug(f"\tExtracting sequences from intervals: {genome_file}")
    annots_per_interval, break_seeds, seed_evidence, inter_seed_dist = extract_seqs_n_annots(merged_bed_files,
                                                                                                   file_type,
                                                                                                   genome_file,
                                                                                                   annotation_file,
                                                                                                   tmp_folder,
                                                                                                   out_path,
                                                                                                   seed_pairs,
                                                                                                   seed_evidence,
                                                                                                   print_seq_out)

    # clean up by removing blast xml output, and merged and seed hit beds from tmp folder:
    os.remove(blast_xml_output)
    tmp_folder_files = os.listdir(tmp_genome_folder)
    for file in tmp_folder_files:
        if str(genome_name).rsplit('/', 1)[1] in file:
            try:
                os.remove(os.path.join(tmp_genome_folder, file))
            except FileNotFoundError:
                pass
    os.rmdir(tmp_genome_folder)
    # tmp_folder_files = os.listdir(tmp_folder)
    # for file in tmp_folder_files:
    #     if '.bed' in file and str(genome_name).rsplit('/')[1] in file:
    #         os.remove(os.path.join(tmp_folder, file))

    return seed_hits, annots_per_interval, genome_name, seed_evidence, break_seeds, inter_seed_dist



import gzip
try:
    from Magphi.exit_with_error import exit_with_error
except ModuleNotFoundError:
    from exit_with_error import exit_with_error

EXIT_INPUT_FILE_ERROR = 3


def check_if_gzip(input_files, file_logger):
    """
    Check if genome files given as input at gzipped, by trying to open them with zgip's open() funciton
    :param
    input_files: A list of names of genomes given as input.
    file_logger: A logger that will write to a logging file
    :return: Boolean that tell if all input files are gzipped.
    """
    file_logger.debug('Check if input genomes are in gzipped state')
    # Make list with Bool for each input genome and evaluate all input genomes for Gzipped state,
    # If genome genome is Gzipped change list index for genome to true in is_input_gzipepd.
    is_input_gzipped = [False] * len(input_files)
    for i, file in enumerate(input_files):
        with gzip.open(file, 'r') as input_file:
            try:
                input_file.readline()
                is_input_gzipped[i] = True
            except OSError:
            # except gzip.BadGzipFile:
                is_input_gzipped[i] = False

    # Check if all inputs are gzipped,
    # if then return True for all inputs being gzipped.
    # Else check if some are gzipped and return that input is mixed
    # Lastly return False if no input is Gzipped.
    if all(is_input_gzipped):
        file_logger.debug('Files were found to be gzipped')
        return True
    elif 0 < sum(is_input_gzipped) < len(is_input_gzipped):
        file_logger.exception('Check input files: Some input files found to be gzipped!\n'
                          'Please check that all of your input files are either Gzipped or not')
        exit_with_error(message='Some input files found to be Gzipped while other were not!',
                        exit_status=EXIT_INPUT_FILE_ERROR)
    else:
        file_logger.debug('Files were not found to be gzipped')
        return False


def check_if_fasta(input_files, file_logger, is_input_gzipped):
    """
    Function to check if input files are Fasta format identified by > in first line
    returns 'fasta' if all files are fasta, exits and logs error if only some are fasta indicating mixed input
    :param input_files: List of input genomes by file name
    :param file_logger: Logger that outputs files to log
    :return: variable: Either 'fasta' if all input genomes are fasta format or None if not
    """
    file_logger.debug('Check if input genomes are fasta format')

    # Construct list to hold if files are fasta
    # Search all files for fasta signature > in first line.
    is_input_fasta = [False]*len(input_files)
    for i, file in enumerate(input_files):
        if is_input_gzipped:
            in_file = gzip.open(file, 'rt')
        else:
            in_file = open(file, 'r')

        # Test first line
        first_line = in_file.readline()
        if '>' in first_line:
            if len(first_line.strip()) > 1:
                is_input_fasta[i] = True
            else:
                in_file.close()
                exit_with_error(
                    f'Fasta file contains no sequence header. This is not allowed please have a look at file: {file}',
                    EXIT_INPUT_FILE_ERROR)

        # Check for new line in remaining lines
        new_line_in_middle = False
        for line in in_file.readlines():
            if new_line_in_middle:
                in_file.close()
                exit_with_error(f'Fasta file contains new line in middle of file. This is not allowed please have a look at file: {file}',
                                EXIT_INPUT_FILE_ERROR)
            # Check for empty line
            if not line.strip():
                new_line_in_middle = True
        in_file.close()


    # Check if all input genomes are fasta, if then return 'fasta'
    # else check if only some input genomes are fasta - if: Give error
    # Else return None
    if all(is_input_fasta):
        file_logger.debug('Files were found to be Fasta files')
        return 'fasta'

    elif any(is_input_fasta):
        file_logger.error('Input files were found to be of mixed type with some being recognised as fasta!\n'
                      'Please make sure that all your input files are only either Fasta or GFF3, not mixed')
        exit_with_error(message='Input files found to be of mixed type! Only some files were recognised as fasta!',
                        exit_status=EXIT_INPUT_FILE_ERROR)

    else:
        file_logger.debug('Files were not found to be Fasta files')
        return None


def check_if_gff(input_files, file_logger, is_input_gzipped):
    """
    Function to check in input genomes are given in a gff3 format with the appended genome.
    :param
    input_files: List of input genomes by file name
    file_logger: Logger that outputs files to log
    :return: Either 'gf' if all input genomes are fasta format or None if not
    """
    ''' Function to check if input files are GFF3 format with an attached genomes identified by the ##FASTA line '''
    file_logger.debug('Check input files: Check if GFF3')
    is_input_gff = [False] * len(input_files)
    for i, file in enumerate(input_files):
        if is_input_gzipped:
            in_file = gzip.open(file, 'rt')
        else:
            in_file = open(file, 'r')

        if '##gff-version 3' in in_file.readline():
            for line in in_file.readlines():
                if '##FASTA' in line:
                    is_input_gff[i] = True
            # Check that a genome has been found in file
            if is_input_gff[i] is False:
                in_file.close()
                print(f'UPS! {file} does seem to be a GFF3 version, but not one that contain the genome '
                      f'following a ##FASTA line - Please check the file')
                file_logger.error('Check input files: Input is GFF but seems to miss ##FASTA line or entire genome')
                exit_with_error(message='Input is GFF but seems to miss ##FASTA line or entire genome',
                                exit_status=EXIT_INPUT_FILE_ERROR)
        in_file.close()

    if all(is_input_gff):
        file_logger.debug('Files were found to be GFF files')
        return 'gff'
    elif any(is_input_gff):
        file_logger.error('Input files were found to be of mixed type with some being recognised as GFF!\n'
                      'Please make sure that all your input files are only either Fasta or GFF3, not mixed')
        exit_with_error(message='Input files found to be of mixed type! Only some files were recognised as GFF!',
                        exit_status=EXIT_INPUT_FILE_ERROR)
    else:
        file_logger.debug('Files were not found to be GFF files')
        return None


def check_inputs(input_files, file_logger):
    """
    Function to run through different checks of input files to determine their state (gzipped) and type (Fasta or GFF)
    :param input_files: List of input genomes by file name
    :param file_logger: Logger that outputs files to log
    :return: Variable for Type, and state of file found by checks.
    """
    ''' Function to check the input files. Will chack if input is GFF, Fasta or exit if filetype is not recognised.
     Will exit program when unrecognised input file is given.'''

    # Add logging step
    file_logger.debug("Check input files: "
                      "Initiating verifying input files are either FASTA or GFF3 with genome attached."
                      " Also checking for gzipped state")

    # Check if files are gzipped
    is_input_gzipped = check_if_gzip(input_files, file_logger)

    # check if input file is fasta
    file_type = check_if_fasta(input_files, file_logger, is_input_gzipped)

    # Check if input files are GFF3
    if file_type is None:
        file_type = check_if_gff(input_files, file_logger, is_input_gzipped)

    # If file_type is still None, exit as input files are not recognized
    if file_type is None:
        file_logger.exception('The given input files could not be recognised as either Fasta or GFF3 with a genome.')
        exit_with_error(message='The given input files could not be recognised as either Fasta or GFF3 with a genome',
                        exit_status=EXIT_INPUT_FILE_ERROR)

    return file_type, is_input_gzipped



import csv
import os


def write_seed_hit_matrix(master_seed_hits, seed_pairs, out_path):
    """
    Function to write the comma delimited file indicating the number of hits for each seed sequence pair
    :param master_seed_hits: Dict with keys being seed sequence pairs, values being a dict with genomes being keys and values being the number of hits for a given seed sequence pair
    :param seed_pairs: Dict of seed sequences that are matched into pairs
    :param out_path: Filepath to the output folder
    :return: None
    """
    with open(os.path.join(out_path, 'contig_hit_matrix.csv'), 'w') as out_file:
        header = ['genome'] + list(seed_pairs.keys())
        writer = csv.DictWriter(out_file, fieldnames=header)

        # Write the field names or header row in file
        writer.writeheader()

        # Write the remaining lines
        keys = list(master_seed_hits.keys())
        keys.sort()
        for line in keys:
            writer.writerow(master_seed_hits[line])

        out_file.close()


def write_annotation_num_matrix(master_annotation_hits, seed_pairs, out_path):
    """
    Function to write a comma separated file giving the number of annotations in regions
    :param master_annotation_hits: Dict with keys being seed sequence pairs, values being a dict with genomes being keys and values being the number of annotations within an extracted region
    :param seed_pairs: Dict of seed sequences that are matched into pairs
    :param out_path: Filepath to the output folder
    :return: None
    """
    with open(os.path.join(out_path, 'annotation_num_matrix.csv'), 'w') as out_file:
        header = ['genome'] + list(seed_pairs.keys())
        writer = csv.DictWriter(out_file, fieldnames=header)

        # Write the field names or header row in file
        writer.writeheader()

        # Write the remaining lines
        keys = list(master_annotation_hits.keys())
        keys.sort()
        for line in keys:
            writer.writerow(master_annotation_hits[line])

        out_file.close()


def write_seed_hit_evidence(master_seed_evidence, seed_pairs, out_path):
    """
    Function to write a comma separated file giving the evidence level for each region identified
    :param master_seed_evidence:
    :param seed_pairs: Dict of seed sequences that are matched into pairs
    :param out_path: Filepath to the output folder
    :return: None
    """
    with open(os.path.join(out_path, 'master_seed_evidence.csv'), 'w') as out_file:
        header = ['genome'] + list(seed_pairs.keys())
        writer = csv.DictWriter(out_file, fieldnames=header)

        # Write the field names or header row in file
        writer.writeheader()

        # Write the remaining lines
        keys = list(master_seed_evidence.keys())
        keys.sort()
        for line in keys:
            writer.writerow(master_seed_evidence[line])

        out_file.close()


def write_inter_seed_dist(master_inter_seed_dist, seed_pairs, out_path):
    """
    Function to write a comma separated file giving the number of nucleotides in regions identified
    :param master_inter_seed_dist:
    :param seed_pairs: Dict of seed sequences that are matched into pairs
    :param out_path: Filepath to the output folder
    :return: None
    """
    with open(os.path.join(out_path, 'inter_seed_distance.csv'), 'w') as out_file:
        header = ['genome'] + list(seed_pairs.keys())
        writer = csv.DictWriter(out_file, fieldnames=header)

        # Write the field names or header row in file
        writer.writeheader()

        # Write the remaining lines
        keys = list(master_inter_seed_dist.keys())
        keys.sort()
        for line in keys:
            writer.writerow(master_inter_seed_dist[line])

        out_file.close()



import subprocess
import warnings
import sys

try:
    from Magphi.exit_with_error import exit_with_error
except ModuleNotFoundError:
    from exit_with_error import exit_with_error
EXIT_DEPENDENCY_ERROR = 3


def check_for_biopython(verbose):
    """
    Function to test presence and version of BioPython
    :param verbose: To print or not to print
    :return: Version of BioPython if present else False
    """
    try:
        import Bio
        if verbose:
            print("Biopython was successfully found")

        if float(Bio.__version__) >= 1.78:
            if verbose:
                print("Biopython version is valid")
            return float(Bio.__version__)

        warnings.warn('Biopython seems to be an untested version! Make sure it is version = 1.78 or above.')
        return False

    except ModuleNotFoundError as exception:
        warnings.warn("Biopython was not found")
        exit_with_error(str(exception), EXIT_DEPENDENCY_ERROR)


def check_for_pybedtools(verbose):
    """
    Function to test presence and version of PyBedtools
    :param verbose: To print or not to print
    :return: Version of PyBedtools if present else False
    """
    try:
        import pybedtools
        if verbose:
            print("pybedtools was successfully found")

        if pybedtools.__version__ >= '0.8.2':
            if verbose:
                print("pybedtools version is valid")
            return pybedtools.__version__

        warnings.warn('pybedtools seems to be an untested version! Make sure it is version = 0.8.2 or above.')
        return False

    except ModuleNotFoundError as exception:
        warnings.warn("pybedtools was not found")
        exit_with_error(str(exception), EXIT_DEPENDENCY_ERROR)


def check_for_bedtools(verbose):
    """
    Function to test presence and version of Bedtools
    :param verbose: To print or not to print
    :return: Version of Bedtools if present else False
    """
    try:
        cmd_return = subprocess.run(['bedtools', '-h'], capture_output=True)

        if cmd_return.returncode == 0:
            if verbose:
                print("Bedtools was successfully found")

            cmd_return = subprocess.run(['bedtools', '-version'], capture_output=True)
            version = cmd_return.stdout.decode().rsplit('v', 1)[-1]
            if version >= '2.29.2':
                if verbose:
                    print("Bedtools version is valid")
                return version

            else:
                warnings.warn('Bedtools seems to be an untested version! Make sure it is version = 2.29.2 or above.')
                return False
        else:
            warnings.warn('The command "bedtools -h" does not seem to give return code = 0')
            sys.exit(EXIT_DEPENDENCY_ERROR)

    except FileNotFoundError as exception:
        warnings.warn('Bedtools could not be found!')
        exit_with_error(str(exception), EXIT_DEPENDENCY_ERROR)


# Samtools
def check_for_samtools(verbose):
    """
    Function to test presence and version of samtools
    :param verbose: To print or not to print
    :return: Version of samtools if present else False
    """
    ''' Check for the presence of samtools and if the version satisfies the required version 1.12, 1.19 or above,
            if this can not be satisfied then warn the use of a missing dependency and exit.
            '''
    try:
        cmd_return = subprocess.run(['samtools', '--help'], capture_output=True)

        if cmd_return.returncode == 0:
            if verbose:
                print("Samtools was successfully found")

            cmd_return = subprocess.run(['samtools', '--version'], capture_output=True)
            # Isolate version tag
            version = cmd_return.stdout.decode().split(' ', 1)[-1]
            version = version.split('\n', 1)[0]

            if version >= '1.11':
                if verbose:
                    print("Samtools version is valid")
                return version

            else:
                warnings.warn('Samtools seems to be an untested version! Make sure it is version = 1.19, 1.12 or above.')
                return False
        else:
            warnings.warn('The command "samtools -h" does not seem to give return code = 0')
            sys.exit(EXIT_DEPENDENCY_ERROR)

    except FileNotFoundError as exception:
        warnings.warn('Samtools could not be found!')
        exit_with_error(str(exception), EXIT_DEPENDENCY_ERROR)


def check_for_blast_plus(verbose=False):
    """
    Function to test presence and version of Blast+
    :param verbose: To print or not to print
    :return: Version of Blast+ if present else False
    """
    ''' Check for the presence of Blast+ and if the version satisfies the required version ,
                if this can not be satisfied then warn the use of a missing dependency and exit. '''

    try:
        cmd_return = subprocess.run(['makeblastdb', '-version'], capture_output=True)

        if cmd_return.returncode == 0:
            if verbose:
                print("Blast+ was successfully found")

            # Isolate version tag
            version = cmd_return.stdout.decode().split('\n')[1]
            version = version.split(' ')[3]
            version = version.rsplit(',', 1)[0]

            if version == '2.6.0' or version >= '2.11.0':
                if verbose:
                    print("Blast+ version is valid")
                return version

            else:
                warnings.warn('Blast+ seems to be an untested version! Make sure it is version = 2.6.0, 2.11 or above.')
                return False
        else:
            warnings.warn('The command "makeblastdb -version" does not seem to give return code = 0')
            sys.exit(EXIT_DEPENDENCY_ERROR)

    except FileNotFoundError as exception:
        warnings.warn('Blast+ could not be found!')
        exit_with_error(str(exception), EXIT_DEPENDENCY_ERROR)


def check_dependencies_for_main(verbose=False):
    """
    Function to be called from main of Magphi to check the dependecies before running
    :param verbose: To print or not to print
    :return: Versions of programs of False if one of them are not present.
    """
    ''' Function to check dependencies when Magphi is executed. If all programs are found their versions are returned,
    if they are not found False is returned.'''
    biopython_present = check_for_biopython(verbose)

    pybedtools_present = check_for_pybedtools(verbose)

    bedtools_present = check_for_bedtools(verbose)

    samtools_present = check_for_samtools(verbose)

    if all([biopython_present, pybedtools_present, bedtools_present, samtools_present]):
        return [biopython_present, pybedtools_present, bedtools_present, samtools_present]
    else:
        return False


def check_dependencies_only():
    """
    Function to check dependencies, when commandline is given '--check' flag
    :return: Nothing
    """
    print('\n------ Checking dependencies for Magphi ------')
    biopython_presence = check_for_biopython(verbose=True)

    pybedtools_presence = check_for_pybedtools(verbose=True)

    bedtools_presence = check_for_bedtools(verbose=True)

    samtools_presence = check_for_samtools(verbose=True)

    check_for_blast_plus(verbose=True)

    print('\n------ Summary of dependency check ------')

    if all([biopython_presence, pybedtools_presence, bedtools_presence, samtools_presence]):
        print("All dependencies were found to be okay! We are ready to go! \n")
    else:
        print("!!! Some dependency does not seem be the correct version, please check the warnings above !!! \n")

    sys.exit(0)


if __name__ == '__main__':
    pass



import argparse
import sys
try:
    from Magphi.check_depencies import check_dependencies_only
except ModuleNotFoundError:
    from check_depencies import check_dependencies_only

EXIT_COMMAND_LINE_ERROR = 2

# def get_commandline_arguments(args, version):
def get_commandline_arguments(args, version):
    """
    Command line interface for Magphi.
    Will print help message and exit upon no input, -help, and check dependencies and exit, upon --check
    :param args: Arguments given on command-line by user
    :param version: Current version of Magphi
    :return: Object containing passed arguments
    """
    # Set up parser
    parser = argparse.ArgumentParser(prog='Magphi',
                                     description='Welcome to Magphi!\n '
                                                 'This program will extract sequences and possible annotations within '
                                                 'a set of seed sequences given as input.')

    # Add the flag for the input genomes
    parser.add_argument('-g',
                        '--input_genomes',
                        help='Give the fasta or gff3 files. (gff3 files should contain the genome fasta sequence)',
                        dest='genomes',
                        required=True,
                        nargs='+',
                        metavar='.fa/.gff')

    # Add the flag for the multi fasta file containing seed sequences
    parser.add_argument('-s',
                        '--input_seeds',
                        help='Give the multi fasta containing the seed sequences to be used for extracting sequnces',
                        dest='seeds',
                        required=True,
                        metavar='multi_fasta_file.fa')

    parser.add_argument('-is',
                        '--include_seeds',
                        help='Argument to include the seeds in the sequence/annotations extracted '
                             '[default: seeds are removed]',
                        dest='include_seeds',
                        required=False,
                        action='store_true',
                        default=False)

    parser.add_argument('-md',
                        '--max_seed_distance',
                        help='The maximum distance with which seeds will be merged.\n'
                             'This can often be set a bit higher than an expected size of a region\n'
                             'If no maximum distance is wanted then set to 0 [default: 50,000bp]',
                        default=50000,
                        required=False,
                        type=int,
                        metavar='int',
                        dest='max_seed_dist')

    output_amount = parser.add_mutually_exclusive_group()
    output_amount.add_argument('-b',
                               '--print_breaks',
                               help='Argument to print outputs when seeds are next to contig breaks'
                                    ' [default: sequences are not printed]',
                               dest='print_breaks',
                               required=False,
                               action='store_true',
                               default=False)

    output_amount.add_argument('-n',
                               '--no_sequences',
                               help='Argument to not print outputs related to annotations or sequences found between seeds'
                                    ' [default: sequences are printed]',
                               dest='no_seqs',
                               required=False,
                               action='store_false',
                               default=True)

    # Add the flag for the output folder
    parser.add_argument('-o',
                        '--output_folder',
                        help='Give path to output folder [default: current folder]',
                        required=False,
                        metavar='path/to/output',
                        default='.',
                        dest='out_path')

    # Add the flag to control max CPUs
    parser.add_argument('-c',
                        '--cpu',
                        help='Give max number of CPUs [default: 1]',
                        required=False,
                        metavar='int',
                        default=1,
                        type=int,
                        dest='cpu')

    logger_level = parser.add_mutually_exclusive_group()
    logger_level.add_argument('-l',
                              '--log',
                              help='Record program progress in for debugging purpose',
                              action='store_true',
                              default=False,
                              required=False)

    logger_level.add_argument('-q',
                              '--quiet',
                              help='Only print warnings',
                              action='store_true',
                              default=False,
                              required=False)

    parser.add_argument('--check',
                              help='Check dependencies for Magphi and exit',
                              dest='dependency_check',
                              action='store_true',
                              default=False,
                              required=False)

    parser.add_argument('-v',
                        '--version',
                        action='version',
                        version=f'Magphi {version}')

    # Check if there are no arguments given or the user ask for the help message
    if len(args) < 1:
        parser.print_help()
        sys.exit(EXIT_COMMAND_LINE_ERROR)
    elif '-help' in args:
        parser.print_help()
        sys.exit(0)
    if '--check' in args:
        check_dependencies_only()

    args = parser.parse_args(args)

    return args



'''
Module      : Main
Description : The main entry point for the program.
Copyright   : (c) Magnus Ganer Jespersen, 11 Oct 2021 
License     : MIT 
Maintainer  : magnus.ganer.j@gmail.com 
Portability : POSIX

The program reads one or more input FASTA files. For each file it computes a
variety of statistics, and then prints a summary of the statistics as output.
'''

import warnings
import os
import time
import logging
from sys import argv
from math import ceil
import pkg_resources # ??
import concurrent.futures

try:
    from Magphi.commandline_interface import get_commandline_arguments
except ModuleNotFoundError:
    from commandline_interface import get_commandline_arguments

try:
    from Magphi.check_depencies import check_dependencies_for_main
except ModuleNotFoundError:
    from check_depencies import check_dependencies_for_main

try:
    from Magphi.exit_with_error import exit_with_error
except ModuleNotFoundError:
    from exit_with_error import exit_with_error

try:
    from Magphi.check_inputs import check_inputs
except ModuleNotFoundError:
    from check_inputs import check_inputs

try:
    from Magphi.split_gff_file import split_gff_files
except ModuleNotFoundError:
    from split_gff_file import split_gff_files

try:
    from Magphi.seed_handling import handle_seeds
except ModuleNotFoundError:
    from seed_handling import handle_seeds

try:
    from Magphi.search_insertion_sites import screen_genome_for_seeds
except ModuleNotFoundError:
    from search_insertion_sites import screen_genome_for_seeds

try:
    from Magphi.wrangle_outputs import partition_outputs, write_paired_seeds
except ModuleNotFoundError:
    from wrangle_outputs import partition_outputs, write_paired_seeds

try:
    from Magphi.write_output_csv import write_seed_hit_matrix, write_annotation_num_matrix, write_seed_hit_evidence, write_inter_seed_dist
except ModuleNotFoundError:
    from write_output_csv import write_seed_hit_matrix, write_annotation_num_matrix, write_seed_hit_evidence, write_inter_seed_dist

# Initial
from argparse import ArgumentParser
from math import floor
import sys
from Bio import SeqIO

EXIT_INPUT_FILE_ERROR = 1
EXIT_COMMAND_LINE_ERROR = 2
EXIT_DEPENDENCY_ERROR = 3
DEFAULT_VERBOSE = False
PROGRAM_NAME = "Magphi"


try:
    PROGRAM_VERSION = pkg_resources.require(PROGRAM_NAME)[0].version
except pkg_resources.DistributionNotFound:
    PROGRAM_VERSION = "undefined_version"


def init_logging(debug_log, quiet, out_path):
    """
    initialise the logging file, and write log statement
    indicating the program has started, and also write out the
    command line from sys.argv
    :param debug_log: Bool indicating if log is a debug log
    :param quiet: Bool indicating if logging should be kept minimal
    :param out_path: Output path for the program, and where log will be placed
    :return: Logger object
    """
    if debug_log:
        level = logging.DEBUG
    elif quiet:
        level = logging.WARNING
    else:
        level = logging.INFO

    # Construct logger logging to file
    file_logger = logging.getLogger(__name__)
    file_logger.setLevel(level)

    formatter = logging.Formatter('[%(asctime)s] %(levelname)s - %(module)s - %(message)s',
                                  datefmt="%Y-%m-%dT%H:%M:%S%z")

    file_handler = logging.FileHandler(os.path.join(out_path, 'Magphi.log'))
    file_handler.setLevel(level)
    file_handler.setFormatter(formatter)
    file_logger.addHandler(file_handler)

    # Log command-line argument and debug line for Magphi start
    file_logger.info(f"command line: {' '.join(argv)}")

    return file_logger


def stream_logging(file_logger):
    """
    Function adding in stream logging following initial logging
    :param file_logger: Logger object
    :return: Logger object with added stream logging
    """
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)

    file_logger.addHandler(stream_handler)

    file_logger.info('Processing started')

    return file_logger


def main():
    """
    This is the main function for running Magphi. Required arguments are genomes and seeds as a multi fasta file
    """
    start_time = time.time()

    # Retrieve the flags given by the user in the commandline
    # TODO - add in argument to change tmp directory to user defined place (scratch which is faster?)
    cmd_args = get_commandline_arguments(argv[1:], PROGRAM_VERSION)

    # Try to construct the output folder and except if it does exist
    try:
        os.mkdir(cmd_args.out_path)
    except FileExistsError:
        warnings.warn("Output folder already exists")
        pass

    "Orchestrate the execution of the program"
    file_logger = init_logging(cmd_args.log, cmd_args.quiet, cmd_args.out_path)

    # Check dependencies for Magphi and logging of versions to file
    dependencies_return = check_dependencies_for_main(verbose=False)
    if dependencies_return:
        file_logger.info("Dependency versions:")
        dependencies = ['Biopython', 'Pybedtools', 'Bedtools', 'Samtools']
        for i in range(0, len(dependencies_return)):
            version = str(dependencies_return[i]).replace("\n", "")
            file_logger.info(f'{dependencies[i]} v.{version}')
    else:
        file_logger.warning("Some dependencies are untested version(s)")

    file_logger = stream_logging(file_logger)

    # Check the input files
    file_type, is_input_gzipped = check_inputs(cmd_args.genomes, file_logger)

    # construct a temporary folder to hold files
    file_logger.debug("Try to construct output folder")
    tmp_folder = os.path.join(cmd_args.out_path, "Magphi_tmp_folder")
    try:
        os.mkdir(tmp_folder)
        file_logger.debug("Output folder construction successful")
    except FileExistsError:
        file_logger.warning("A temporary folder already exists at the given output location. "
                                     "Most likely from an incomplete analysis")

    # Read in and combine seeds into pairs
    file_logger.debug("Start handling of input seed sequences")
    seed_pairs = handle_seeds(cmd_args.seeds, file_logger)

    print_seq_out = 'output' if cmd_args.no_seqs else 'None'
    if print_seq_out == 'output':
        print_seq_out = 'All' if cmd_args.print_breaks else print_seq_out

    # Construct master dict to hold the returned information from seeds
    master_seed_hits = {}
    master_annotation_hits = {}
    master_seed_evidence = {}
    seeds_w_breaks = seed_pairs.copy()
    master_inter_seed_dist = {}

    # num_genomes = len(genomes)
    num_genomes = len(cmd_args.genomes)
    # Calculate when to log progress:
    progress_num = num_genomes / 10
    progress_num = progress_num if progress_num > 1 else 1
    progress_num = ceil(progress_num)

    file_logger.info(f'{num_genomes} input files to be processed, starting now!')
    genomes_processed = 0
    with concurrent.futures.ThreadPoolExecutor(max_workers=cmd_args.cpu) as executor:
        results = [executor.submit(screen_genome_for_seeds, cmd_args.genomes[i], seed_pairs, cmd_args.seeds,
                                   tmp_folder, cmd_args.include_seeds, file_type,
                                   cmd_args.out_path, cmd_args.max_seed_dist, file_logger, is_input_gzipped, print_seq_out)
                   for i, genome in enumerate(cmd_args.genomes)]

        for f in concurrent.futures.as_completed(results):
            genomes_processed += 1

            if genomes_processed % progress_num == 0 or genomes_processed == 1:
                file_logger.info(f'\tFile number {genomes_processed} has been processed')

            seed_hits, annots_per_interval, genome_name, seed_evidence, break_seeds, inter_seed_dist = f.result()

            # Polish the genome name for the output dict, and possible .fasta and .gff file extensions:
            genome_name = genome_name.rsplit('/', 1)[-1]
            genome_name = genome_name.rsplit('.gff', 1)[0]
            genome_name = genome_name.rsplit('.fasta', 1)[0]
            file_logger.debug(f'\t\tCurrently handling results from: {genome_name}')

            # Update the master dicts with information from current run.
            master_annotation_hits[genome_name] = annots_per_interval
            # master_annotation_hits[genome_name] = annots_per_interval
            master_seed_hits[genome_name] = seed_hits
            # master_seed_hits[genome_name] = seed_hits
            master_seed_evidence[genome_name] = seed_evidence
            master_inter_seed_dist[genome_name] = inter_seed_dist

            # Add the genome to the dict to be writen out
            master_seed_hits[genome_name]['genome'] = genome_name
            master_annotation_hits[genome_name]['genome'] = genome_name
            master_seed_evidence[genome_name]['genome'] = genome_name
            master_inter_seed_dist[genome_name]['genome'] = genome_name

            # Check if there are any seeds returned as being neighbours to contig breaks.
            if len(break_seeds.keys()) > 0:
                seeds_w_breaks.update(break_seeds)

    # Partition output files into their seed set of origin.
    partition_outputs(seed_pairs, cmd_args.out_path, file_logger)

    file_logger.debug('Start writing output files')
    file_logger.debug('\tStart writing hit matrix file')
    write_seed_hit_matrix(master_seed_hits, seed_pairs, cmd_args.out_path)
    if file_type == 'gff':
        file_logger.debug('\tStart writing hit extracted annotation file')
        write_annotation_num_matrix(master_annotation_hits, seeds_w_breaks, cmd_args.out_path)
    file_logger.debug('\tStart writing evidence for seed sequence pair file')
    write_seed_hit_evidence(master_seed_evidence, seed_pairs, cmd_args.out_path)
    file_logger.debug('\tStart writing distance between seed sequence pair file')
    write_inter_seed_dist(master_inter_seed_dist, seeds_w_breaks, cmd_args.out_path)
    file_logger.debug('\tStart writing seed sequence pairing file')
    write_paired_seeds(seed_pairs, cmd_args.out_path)

    # log quick stats for the evidence levels
    file_logger.debug("Remove temporary folder")
    os.rmdir(tmp_folder)

    time_to_finish = time.time() - start_time
    time_to_finish = int(round(time_to_finish, 0))
    file_logger.info(f"Magphi completed in: {time_to_finish//60} minutes {time_to_finish%60} Seconds")


if __name__ == '__main__':
    main()


import os
import warnings


def partition_outputs(seed_pairs, out_path, file_logger):
    """
    Function to distribute output files into appropriate output folders
    :param seed_pairs: Name of seed sequence pairs in a list
    :param out_path: Path to the output folder
    :param file_logger: Logger that outputs files to log
    :return: None
    """
    file_logger.debug('Partitioning output files into seed folders')
    # Go through each seed sequence pair, construct sub output folder and populate with output files.
    for seed in seed_pairs:
        # List all output files in output folder
        out_folder_content = os.listdir(out_path)
        out_folder_content_paths = [os.path.join(out_path, file) for file in out_folder_content]

        # Construct folder for current seed to contain fasta and gff files for that seed
        seed_out_path = os.path.join(out_path, seed)
        try:
            os.mkdir(seed_out_path)
        except FileExistsError:
            file_logger.warning(f"Output folder for seed sequence pair: '{seed}' already exists")

        # find and move files into folder
        for i, file in enumerate(out_folder_content_paths):

            # Check that the marker of a output file is present (--)
            if '--' in file and ('.gff' in file or '.fasta' in file):
                # Isolate the seed from the file name
                seed_end = file.rsplit('--', 1)[-1]
                file_seed = seed_end.rsplit('.', 1)[0]

                # Check if break is present, if then remove the break and then narrow down the seed the file belongs to.
                if 'break' in file_seed:
                    file_seed = file_seed.rsplit('_', 1)[0]
                    seed_match = [file_seed != seed_name for seed_name in seed_pairs]

                    round_count = 1
                    while all(seed_match):
                        file_seed = file_seed[:-1]
                        seed_match = [file_seed != seed_name for seed_name in seed_pairs]

                        round_count += 1

                        if round_count == 100:
                            warnings.warn(f"Problem in placing outfile: '{file}' in an appropriate output folder for a seed.\n"
                                          f"Please report this!")
                            file_logger.warning(f"Problem in placing outfile: '{file}' in an appropriate output folder for a seed.\n"
                                                f"Please report this!")
                            continue

                # Check if a file in the output folder contain the seed name and is a fasta or gff file.
                # if seed == file_seed and (f'{seed}_' in seed_end or f'{seed}.' in seed_end):
                if seed == file_seed:
                    # Replace the double hyphen with a single.
                    file_new = out_folder_content[i].replace('--', '-')
                    # Move the file into the seed folder
                    os.rename(file, os.path.join(seed_out_path, file_new))


def write_paired_seeds(seed_pairs, out_path):
    """
    Function to write a tab delimited file giving the pairing for seed seqeunces
    :param seed_pairs: Dict of seed sequences that are matched into pairs
    :param out_path: Filepath to the output folder
    :return: None
    """
    with open(os.path.join(out_path, 'seed_pairing.tsv'), 'w') as seed_file:
        for seed_key in seed_pairs.keys():
            # Fetch the seeds
            seeds = seed_pairs[seed_key]
            # Write the line
            seed_file.write(f'{seed_key}\t{seeds[0]}\t{seeds[1]}\n')
        seed_file.close()



#!/usr/bin/env python

from setuptools import setup

LONG_DESCRIPTION = \
'''The program extracts regions of interest from Fasta or Genome Feature Format (GFF) genomes.
This is done given a set of seed sequences given as nucleotide strings in a multi-line fasta file.  
The program can output fasta and GFF outputs or regions, and will giv multiple outputs around regions and their evidence.

The program takes in multiple fasta or GFF files at a single time and can also take multiple seed sequence pairs.'''


setup(
    name='Magphi',
    version='1.0.1',
    author='Magnus Ganer Jespersen',
    author_email='magnus.ganer.j@gmail.com',
    packages=['Magphi'],
    package_dir={'Magphi': 'Magphi'},
    entry_points={
        'console_scripts': ['Magphi = Magphi.__main__:main']
    },
    url='https://github.com/milnus/Magphi',
    license='MIT license',
    description=('A bioinformatics tool allowing for examnination and extraction of genomic features using seed sequences.'),
    long_description=LONG_DESCRIPTION,
    install_requires=['biopython==1.79',
                      'pybedtools'],
    keywords=['Genomic', 'extraction', 'bacteria', 'prokaryotes', 'bioinformatics'],
    classifiers=[
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Development Status :: 4 - Beta']
)



import os
import concurrent.futures
import gzip


def split_single_gff(gff, tmp_folder, is_input_gzipped):
    """
    Function that splits an input GFF3 files into two files, one containing the annotations and another the genome
    :param gff: Filepath to the input GFF file
    :param tmp_folder: Filepath to the temporary folder constructed by Magphi
    :param is_input_gzipped: Bool to indicate if the input GFF file is gzipped

    :return tmp_fasta_name: Filepath to the temporary FASTA file containing the genome sequence from the input
    :return tmp_gff_name: Filepath to the temporary GFF file containing the annotations from the input
    """
    # Trim the name/path of the input file
    base_name = gff.split('/')[-1]
    base_name = base_name.rsplit('.gz', 1)[0]
    base_name = base_name.rsplit('.', 1)[0]

    # Add the path to the temporary folder
    outfolder_basename = os.path.join(tmp_folder, base_name)

    # Construct the names of the temporary files
    tmp_fasta_name = f'{outfolder_basename}_tmp.fasta'
    tmp_gff_name = f'{outfolder_basename}_tmp.gff'

    # Initialise the temporary files for the genome and annotations
    tmp_gff = open(tmp_gff_name, 'w')
    tmp_fasta = open(tmp_fasta_name, 'w')

    # Open the input gff file
    if is_input_gzipped:
        gff_file = gzip.open(gff, 'rt')
    else:
        gff_file = open(gff, 'r')
    # Indicate whether fasta file is found
    fasta_found = False

    # Go through all line of gff file
    for line in gff_file.readlines():
        # Check if genome has been reach if, then skip line
        if '##FASTA' in line:
            fasta_found = True
            continue

        # Check if genome has been reach if, then record line in genome,
        # else then record the as annotation line
        if fasta_found:
            tmp_fasta.write(line)
        else:
            tmp_gff.write(line)

    # Close files before opening the next
    gff_file.close()
    tmp_gff.close()
    tmp_fasta.close()

    return tmp_fasta_name, tmp_gff_name


def split_gff_files(input_files, tmp_folder, is_input_gzipped):
    """ Function to go through each input GFF3 file and split it into temporary files containing the gff annotations
    and the genome, to make them compatible with BLAST and bedtools"""

    # initialise lists to hold the name of the fasta and annotations files
    genome_files = []
    annotation_files = []

    with concurrent.futures.ThreadPoolExecutor() as executor:
        results = [executor.submit(split_single_gff, gff, tmp_folder, is_input_gzipped) for gff in input_files]

        for f in concurrent.futures.as_completed(results):
            tmp_fasta, tmp_gff = f.result()

            genome_files.append(tmp_fasta)
            annotation_files.append(tmp_gff)

    return genome_files, annotation_files



import os
import sys
try:
    from Magphi.exit_with_error import exit_with_error
except ModuleNotFoundError:
    from exit_with_error import exit_with_error

EXIT_INPUT_FILE_ERROR = 3


def check_number_n_names_of_seeds(seed_file, file_logger):
    """
    Function to count and report on the number of seed sequences and pairs of seeds expected
    Also record all names of seeds sequences in a list
    :param
    seed_file: File name of multi fasta file containing seed sequences
    file_logger: Logger that outputs files to log
    :return: Will throw error is number of seed sequences is odd
    or if two seeds have an identical name,
    else returns the number of seeds and the names in a list
    """
    file_logger.debug("Initiating count of seed sequences")
    # Initiate seed counter
    num_seeds = 0
    seed_list = []

    # Go through multi fasta file of seed sequences
    with open(seed_file, 'r') as seeds:
        for line in seeds.readlines():
            # Check if a new seed sequence has been reached, if then increment seed sequence count
            # and add the name to the list of seeds
            if ">" in line:
                num_seeds += 1

                line = line.strip('\n')
                seed_name = line[1:]

                if seed_name in seed_list:
                    file_logger.exception(
                        "Duplicate seed sequence names were identified! This is not allowed and should be changed")
                    exit_with_error(
                        message="Duplicate seed sequence names were identified! This is not allowed and should be changed",
                        exit_status=EXIT_INPUT_FILE_ERROR)

                seed_list.append(seed_name)

    if num_seeds % 2 != 0:
        file_logger.exception("The number of seed sequences given is not even! This means that not all seed sequences can be given a mate\n"
              "If you have a seed sequence that should be used twice it should be in the file twice.")
        exit_with_error(message="The number of seed sequences given is not even! This means that not all seed sequences can be given a mate\n"
                                "If you have a seed sequence that should be used twice it should be in the file twice.",
                                        exit_status=EXIT_INPUT_FILE_ERROR)
    else:
        file_logger.debug("Number of seed sequences given is equal")
        return num_seeds, seed_list


def construct_pair_seeds(seed_names, file_logger):
    """
    Function to pair names of seed sequences into pairs of two.
    :param seed_names: Names of seed sequences from input file
    :param file_logger: Logger that outputs files to log
    :return: A dict with key names being seed sequence pair name and values the name of seed sequences
    """
    file_logger.debug("Initiate pairing of seed sequence names into pairs")
    # Initiate dict that contain the basename of a seed pair as the key and seed names as the values
    seed_pairs = {}

    # Loop through all seed sequence names to find pairs of seed sequences
    n_loops = 0
    while len(seed_names) > 0 and n_loops != 1000:
        basename_lengths = []
        # Choose and extract fist seed in list
        chosen_seed = seed_names.pop(0)

        # Go through seed sequences and find the length of the base name compared to the chosen
        for i, seed in enumerate(seed_names):
            # Compare chosen seed to a possible mate and extract the length of base/common name
            basename_lengths.append(len(os.path.commonprefix([chosen_seed, seed])))

        # Find the seed(s) with the longest base name
        longest_mates_index = [mate_index for mate_index, mate in enumerate(basename_lengths) if mate == max(basename_lengths)]

        # Check if only one seed sequence is found to have the longest base/common name with chosen seed sequence,
        #   if then record the pair,
        #   if not then return the chosen one to the list a restart
        if len(longest_mates_index) == 1:
            # Get the best match
            chosen_mate = seed_names[longest_mates_index[0]]

            # Construct the pair
            seed_pair = [chosen_mate, chosen_seed]

            # Find the common name for pair
            common_name = os.path.commonprefix(seed_pair)

            if common_name[-1] == '_':
                common_name = common_name[:-1]

            # Insert pair into seed dict
            seed_pairs[common_name] = seed_pair

            # Delete seed chosen as mate from list of seeds
            del seed_names[longest_mates_index[0]]
        else:
            seed_names.append(chosen_seed)

        n_loops += 1

    if n_loops == 1000:
        file_logger.exception(f'Seed sequence pairing failed due to more than 1000 rounds of pairing tested. seeds remaining to be paired: seed_names = {seed_names}')
        exit_with_error(message=f'Seed sequence pairing failed due to more than 1000 rounds of pairing tested. seeds remaining to be paired: seed_names = {seed_names}',
                                        exit_status=EXIT_INPUT_FILE_ERROR)

    return seed_pairs


def handle_seeds(seed_file, file_logger):
    """
    Function to be called in main that will check, extract info from, and pair seed sequences into pairs
    returns the Seed sequence pairs and a dict with name and sequences of seeds
    :param seed_file: File name of multi sequence fasta file containing the seed sequences
    :param file_logger: Logger that outputs files to log
    :return: pairs of seed sequences
    """

    num_seeds, seed_list = check_number_n_names_of_seeds(seed_file, file_logger)

    seed_pairs = construct_pair_seeds(seed_list, file_logger)

    file_logger.debug(f'{num_seeds} seeds found in seed file. - '
                      f'This leads to {len(seed_pairs)} pairs of seeds.')

    return seed_pairs
